{************************************************}
{                                                }
{   Turbo Pascal 7.0                             }
{   FreePacal                                    }
{   Delphi 4                                     }
{   Calendar                                     }
{   Copyright (c) 1995,96,97 by Alexs            }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}
unit ASCIITab;
interface

uses Objects, App, Views, Drivers, Register;

type
  PASCIITable = ^TASCIITable;
  TASCIITable = object(TView)
    constructor Init(var Bounds: TRect);
    procedure Draw; virtual;
    procedure HandleEvent(var Event:TEvent); virtual;
  end;

  PASCIIReport = ^TASCIIReport;
  TASCIIReport = object(TView)
    ASCIIChar: LongInt;
    constructor Init(var Bounds: TRect);
    constructor Load(var S: TStream);
    procedure Draw; virtual;
    procedure HandleEvent(var Event:TEvent); virtual;
    procedure Store(var S: TStream);
  end;

  PASCIIChart = ^TASCIIChart;
  TASCIIChart = object(TWindow)
    constructor Init;
  end;

const
  AsciiTableCommandBase = 21100;

  RASCIITable: TStreamRec = (
     ObjType: rcASCIITable;
     VmtLink: Ofs(TypeOf(TASCIITable)^);
     Load:    @TASCIITable.Load;
     Store:   @TASCIITable.Store
  );
  RASCIIReport: TStreamRec = (
     ObjType: rcASCIIReport;
     VmtLink: Ofs(TypeOf(TASCIIReport)^);
     Load:    @TASCIIReport.Load;
     Store:   @TASCIIReport.Store
  );
  RASCIIChart: TStreamRec = (
     ObjType: rcASCIIChart;
     VmtLink: Ofs(TypeOf(TASCIIChart)^);
     Load:    @TASCIIChart.Load;
     Store:   @TASCIIChart.Store
  );

procedure RegisterASCIITab;

implementation

const
  cmCharacterFocused = 0;

constructor TASCIITable.Init(var Bounds: TRect);
begin
  inherited Init(Bounds);
  EventMask := EventMask or evKeyBoard;
  State := State or sfCursorVis;
  Options := Options or (ofSelectable + ofFirstClick + ofVersion20 + ofFramed);
  BlockCursor;
  ComponentType:=ctASCIITable;
end;


procedure TASCIITable.Draw;
var
  Buf: TDrawBuffer;
  X, Y: Integer;
  Color: Byte;
begin
  Color := GetColor(6);
  for Y := 0 to Size.Y - 1 do
  begin
    MoveChar(Buf, ' ', Color, Size.X);
    for X := 0 to Size.X - 1 do
      MoveChar(Buf[x], Chr(32 * y + x), Color, 1);
    WriteLine(0, y, Size.X, 1, Buf);
  end;
end;

procedure TASCIITable.HandleEvent(var Event:TEvent);
var
  CurrentSpot: TPoint;

procedure CharFocused;
begin
  Message(Owner, evBroadcast, AsciiTableCommandBase + cmCharacterFocused,
    Pointer(Cursor.X + 32 * Cursor.Y));
end;

begin
  inherited HandleEvent(Event);
  if Event.What = evMouseDown then
  begin
    repeat
      if MouseInView(Event.Where) then
      begin
        MakeLocal(Event.Where, CurrentSpot);
        SetCursor(CurrentSpot.X, CurrentSpot.Y);
        CharFocused;
      end;
    until not MouseEvent(Event, evMouseMove);
    ClearEvent(Event);
  end
  else if Event.What = evKeyDown then
    with Cursor do begin
      case Event.KeyCode of
        kbHome: SetCursor(0,0);
        kbEnd: SetCursor(Size.X - 1, Size.Y - 1);
        kbUp: if Y > 0 then SetCursor(X, Y - 1);
        kbDown: if Y < Size.Y - 1 then SetCursor(X, Y + 1);
        kbLeft: if X > 0 then SetCursor(X - 1, Y);
        kbRight: if X < Size.X - 1 then SetCursor(X + 1, Y);
      else
        SetCursor(ord(Event.CharCode) mod 32, ord(Event.CharCode) div 32);
      end;
      CharFocused;
      ClearEvent(Event);
    end;
end;

{ TASCIIReport }

constructor TASCIIReport.Init(var Bounds: TRect);
begin
  inherited Init(Bounds);
  EventMask := EventMask or evBroadcast;
  Options := Options or ofFramed;
  ComponentType:=ctASCIIReport;
end;

constructor TASCIIReport.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(ASCIIChar, SizeOf(ASCIIChar));
end;

procedure TASCIIReport.Draw;
var
  Ch: LongInt;
  Color: Byte;
  Buf: TDrawBuffer;
  TempStr: string;
begin
  FormatStr(TempStr, '  Char: %c Decimal: %0#%3d Hex: %0#%02x  ', ASCIIChar);
  WriteStr(0, 0, TempStr, 6);
end;

procedure TASCIIReport.HandleEvent(var Event: TEvent);
var
  ASCIITable: PASCIITable;
begin
  inherited HandleEvent(Event);
  if Event.What = evBroadcast then
    if Event.Command = AsciiTableCommandBase + cmCharacterFocused then
    begin
      ASCIIChar := Event.InfoLong;
      DrawView;
    end;
end;

procedure TASCIIReport.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(ASCIIChar, SizeOf(ASCIIChar));
end;

  { TASCIIChart }
constructor TASCIIChart.Init;
var
  R: TRect;
  Control: PVIew;
begin
  R.Assign(0, 0, 34, 12);
  TWindow.Init(R, 'ASCII Chart', wnNoNumber);
  Flags := Flags and not (wfGrow + wfZoom);
  Palette := wpGrayWindow;

  R.Grow(-1,-1);
  R.A.Y := R.B.Y - 1;
  Insert(New(PASCIIReport, Init(R)));
  GetExtent(R);
  R.Grow(-1,-1);
  R.B.Y := R.B.Y - 2;
  Control := New(PASCIITable, Init(R));
  Insert(Control);
  Control^.Select;
  ComponentType:=ctASCIIChart;
end;

procedure RegisterASCIITab;
begin
  RegisterType(RASCIITable);
  RegisterType(RASCIIReport);
  RegisterType(RASCIIChart);
end;

end.
