{$O+}
{$I Define.inc}
unit CmdNamer;

interface
uses Strings, Objects, Views, App, Dialogs, ColorSel, AddCntrl, DBGrids,
     HelpFile, DataCtrl, AboutPn, Validate, El_Lists, Register;
type
  PCommandCollection = ^TCommandCollection;
  TCommandCollection = object(TSortedCollection)
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    procedure FreeItem(P: Pointer); virtual;
    function KeyOf(P: Pointer): Pointer; virtual;
    function GetItem(var S: TStream): Pointer; virtual;
    procedure PutItem(var S: TStream; Item: Pointer); virtual;
  end;

  PCommandNameCollection = ^TCommandNameCollection;
  TCommandNameCollection = object(TCommandCollection)
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    function KeyOf(P: Pointer): Pointer; virtual;
  end;

type
  PCommandLookupValidator = ^TCommandLookupValidator;
  TCommandLookupValidator = object(TLookupValidator)
    CommandName:PCommandNameCollection;
    constructor Init;
    constructor Load(var S: TStream);
    procedure Error; virtual;
    function Lookup(const S: string): Boolean; virtual;
  end;


{ Registers all the built in Turbo Vision commands with the command
  namer }
procedure BuiltInCommandNames;

{ Returns the name of the given command }
function CommandName(Command: Word): String;

{ Registers the given command with the command namer }
procedure NameCommand(Command: Word; const Name: String);
{ Default commands }

procedure RegisterCmdNamer;
type
  TDefName = record
    Name: PChar;
    Command: Word;
  end;

const
  DefCommands: array[0..59+4] of TDefName = (
    (Name: 'cmValid'; Command: cmValid),
    (Name: 'cmQuit'; Command: cmQuit),
    (Name: 'cmError'; Command: cmError),
    (Name: 'cmMenu'; Command: cmMenu),
    (Name: 'cmClose'; Command: cmClose),
    (Name: 'cmZoom'; Command: cmZoom),
    (Name: 'cmResize'; Command: cmResize),
    (Name: 'cmNext'; Command: cmNext),
    (Name: 'cmPrev'; Command: cmPrev),
    (Name: 'cmHelp'; Command: cmHelp),
    (Name: 'cmOK'; Command: cmOK),
    (Name: 'cmCancel'; Command: cmCancel),
    (Name: 'cmYes'; Command: cmYes),
    (Name: 'cmNo'; Command: cmNo),
    (Name: 'cmDefault'; Command: cmDefault),
    (Name: 'cmCut'; Command: cmCut),
    (Name: 'cmCopy'; Command: cmCopy),
    (Name: 'cmPaste'; Command: cmPaste),
    (Name: 'cmUndo'; Command: cmUndo),
    (Name: 'cmClear'; Command: cmClear),
    (Name: 'cmTile'; Command: cmTile),
    (Name: 'cmCascade'; Command: cmCascade),
    (Name: 'cmNew'; Command: cmNew),
    (Name: 'cmOpen'; Command: cmOpen),
    (Name: 'cmSave'; Command: cmSave),
    (Name: 'cmSaveAs'; Command: cmSaveAs),
    (Name: 'cmSaveAll'; Command: cmSaveAll),
    (Name: 'cmChangeDir'; Command: cmChangeDir),
    (Name: 'cmDosShell'; Command: cmDosShell),
    (Name: 'cmCloseAll'; Command: cmCloseAll),
    (Name: 'cmReceivedFocus'; Command: cmReceivedFocus),
    (Name: 'cmReleasedFocus'; Command: cmReleasedFocus),
    (Name: 'cmCommandSetChanged'; Command: cmCommandSetChanged),
    (Name: 'cmScrollBarChanged'; Command: cmScrollBarChanged),
    (Name: 'cmScrollBarClicked'; Command: cmScrollBarClicked),
    (Name: 'cmSelectWindowNum'; Command: cmSelectWindowNum),
    (Name: 'cmListItemSelected'; Command: cmListItemSelected),

    (Name: 'cmColorForegroundChanged'; Command:cmColorForegroundChanged),
    (Name: 'cmColorBackgroundChanged'; Command:cmColorBackgroundChanged),
    (Name: 'cmColorSet'; Command:cmColorSet),
    (Name: 'cmNewColorItem'; Command:cmNewColorItem),
    (Name: 'cmNewColorIndex'; Command:cmNewColorIndex),
    (Name: 'cmSaveColorIndex'; Command:cmSaveColorIndex),
    (Name: 'cmDBGreedSelect'; Command:cmDBGridSelect),
    (Name: 'cmDataScrollerClicked'; Command:cmDataScrollerClicked),
    (Name: 'cmDataScrollerChanged'; Command:cmDataScrollerChanged),
    {AddCntrl}
    (Name: 'cmNewElementItem'; Command:cmNewElementItem),
    (Name: 'cmColorTextChanged'; Command:cmColorTextChanged),
    {HelpFile}
    (Name: 'cmPrevScr';  Command:cmPrevScr),
    (Name: 'cmFindHelpWindow'; Command:cmFindHelpWindow),
     {DataCtrl}
    (Name: 'cmPrev'; Command     :cmPrevRec),
    (Name: 'cmNext'; Command     :cmNextRec),
    (Name: 'cmFirst'; Command    :cmFirstRec),
    (Name: 'cmLast'; Command     :cmLastRec),
    (Name: 'cmMoveTo'; Command   :cmMoveBy),
    (Name: 'cmUpDataDS'; Command :cmUpDataDS),
    (Name: 'cmNewTable'; Command :cmNewTable),
    (Name: 'cmDeleteRec'; Command:cmDeleteRec),
    (Name: 'cmRecalRec'; Command :cmRecalRec),
    (Name: 'cmEditRec'; Command  :cmEditRec),
    (Name: 'cmAppendRec'; Command:cmAppendRec),
    (Name: 'cmPostRec'; Command  :cmPostRec),
    (Name: 'cmCancelRec'; Command:cmCancelRec),

    {AboutPn}
    (Name: 'cmIdle'; Command: cmIdle));


var
  CommandDB: PCommandCollection;
  CommandNameDB: PCommandNameCollection;

const
  RCommandCollection: TStreamRec = (
     ObjType: 2018;
     VmtLink: Ofs(TypeOf(TCommandCollection)^);
     Load:    @TCommandCollection.Load;
     Store:   @TCommandCollection.Store
  );
const
  RCommandNameCollection: TStreamRec = (
     ObjType: 2020;
     VmtLink: Ofs(TypeOf(TCommandNameCollection)^);
     Load:    @TCommandNameCollection.Load;
     Store:   @TCommandNameCollection.Store
  );
const
  RCommandLookupValidator: TStreamRec = (
     ObjType: 2021;
     VmtLink: Ofs(TypeOf(TCommandLookupValidator)^);
     Load:    @TCommandLookupValidator.Load;
     Store:   @TCommandLookupValidator.Store
  );
implementation
uses MsgBox, Ext_Str;

{ Command Database }

type
  PCommandItem = ^TCommandItem;
  TCommandItem = record
    Command: Word;
    Name: PString;
  end;


function TCommandCollection.Compare(Key1, Key2: Pointer): Integer;
begin
  Compare := LongInt(Key1) - LongInt(Key2);
end;

procedure TCommandCollection.FreeItem(P: Pointer);
begin
  DisposeStr(PCommandItem(P)^.Name);
  Dispose(PCommandItem(P));
end;

function TCommandCollection.GetItem(var S: TStream): Pointer;
var
   P:PCommandItem;
begin
  New(P);
  S.Read(P^.Command,SizeOf(word));
  P^.Name:=S.ReadStr;
  GetItem:=P;
end;

procedure TCommandCollection.PutItem(var S: TStream; Item: Pointer);
var
   P:PCommandItem absolute Item;
begin
  S.Write(P^.Command,SizeOf(word));
  S.WriteStr(P^.Name);
end;

function TCommandCollection.KeyOf(P: Pointer): Pointer;
begin
  KeyOf := Pointer(PCommandItem(P)^.Command);
end;


function TCommandNameCollection.Compare(Key1, Key2: Pointer): Integer;
var Str1:PString absolute Key1;
    Str2:PString absolute Key2;
begin
  if StUpCase(Str1^) < StUpCase(Str2^) then Compare:=-1
  else if StUpCase(Str1^) > StUpCase(Str2^) then Compare:=1
  else Compare:=0;
end;

function TCommandNameCollection.KeyOf(P: Pointer): Pointer;
var AP:PCommandItem absolute P;
begin
  KeyOf:=AP^.Name;
end;

procedure BuiltInCommandNames;
var
  I: Integer;
begin
  for I := Low(DefCommands) to High(DefCommands) do
    NameCommand(DefCommands[I].Command, StrPas(DefCommands[I].Name));
end;

procedure NameCommand(Command: Word; const Name: String);
var
  P,P1: PCommandItem;
begin
  New(P);
  P^.Command := Command;
  P^.Name := NewStr(Name);
  CommandDB^.Insert(P);
  New(P1);
  P1^.Command := Command;
  P1^.Name := NewStr(Name);
  CommandNameDB^.Insert(P1);
end;

function CommandName(Command: Word): String;
var
  I: Integer;
  S: String;
begin
  if CommandDB^.Search(Pointer(Command), I) then
    CommandName := PCommandItem(CommandDB^.At(I))^.Name^
  else
  begin
    Str(Command, S);
    CommandName := 'unknown ' + S;
  end;
end;

{ TCommandLookupValidator }
constructor TCommandLookupValidator.Init;
begin
  inherited Init;
  CommandName:=CommandNameDB
end;

constructor TCommandLookupValidator.Load(var S: TStream);
begin
  inherited Load(S);
  CommandName := CommandNameDB;
end;

procedure TCommandLookupValidator.Error;
begin
  MessageBox('Input not in valid command', nil, mfError + mfOKButton);
end;

function TCommandLookupValidator.Lookup(const S: string): Boolean;
var
  Index: Integer;
  Str: PString;
begin
  asm
        LES     DI,S
        MOV     Str.Word[0], DI
        MOV     Str.Word[2], ES
  end;
  Lookup := False;
  if CommandName <> nil then
    Lookup := CommandName^.Search(Str, Index);
end;

procedure RegisterCmdNamer;
begin
  RegisterType(RCommandCollection);
  RegisterType(RCommandNameCollection);
  RegisterType(RCommandLookupValidator);
end;

end.
