{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       FreePacal                                       }
{       Delphi 4                                        }
{       Turbo Vision Unit                               }
{       Extended visual component                       }
{                                                       }
{       Copyright (c) 1995-98 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}
unit El_Lists;
interface

uses Objects, Views, Drivers, Register;

type
  { TElementItem }
  PElementItem = ^TElementItem;
  TElementItem = record
    Name: PString;
    Next: PElementItem;
  end;

  { TElementGroup }
  PElementGroup = ^TElementGroup;
  TElementGroup = record
    Name:  PString;
    Index: Byte;
    Items: PElementItem;
    Next:  PElementGroup;
  end;

type
  { TElementGroupList }

  PElementGroupList = ^TElementGroupList;
  TElementGroupList = object(TListViewer)
    Groups: PElementGroup;
    constructor Init(var Bounds: TRect; AScrollBar: PScrollBar;
      AGroups: PElementGroup);
    constructor Load(var S: TStream);
    destructor Done; virtual;
    procedure Awaken; virtual;
    procedure FocusItem(Item: Integer); virtual;
    function GetText(Item: Integer; MaxLen: Integer): String; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Store(var S: TStream);
    procedure SetGroupIndex(GroupNum, ItemNum: Byte);
    function GetGroup(GroupNum: Byte): PElementGroup;
    function GetNumGroups: byte;
    function GetGroupIndex(GroupNum: Byte): byte;
  end;

type

  { TElementItemList }

  PElementItemList = ^TElementItemList;
  TElementItemList = object(TListViewer)
    Items: PElementItem;
    constructor Init(var Bounds: TRect; AScrollBar: PScrollBar;
      AItems: PElementItem);
    procedure FocusItem(Item: Integer); virtual;
    function GetText(Item: Integer; MaxLen: Integer): String; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    function DataSize:Word;virtual;
    procedure GetData(var Rec);virtual;
    procedure SetData(var Rec);virtual;
  end;

function ElementItem(const Name: String; Next: PElementItem): PElementItem;
function ElementGroup(const Name: String; Items: PElementItem;
  Next: PElementGroup): PElementGroup;

const
  RElementGroupList: TStreamRec = (
     ObjType: rcElementGroupList;
     VmtLink: Ofs(TypeOf(TElementGroupList)^);
     Load:    @TElementGroupList.Load;
     Store:   @TElementGroupList.Store
  );

const
  RElementItemList: TStreamRec = (
     ObjType: rcElementItemList;
     VmtLink: Ofs(TypeOf(TElementItemList)^);
     Load:    @TElementItemList.Load;
     Store:   @TElementItemList.Store
  );

procedure RegisterEl_Lists;

implementation

{ TElementGroupList }

constructor TElementGroupList.Init(var Bounds: TRect; AScrollBar: PScrollBar;
  AGroups: PElementGroup);
var
  I: Integer;
begin
  TListViewer.Init(Bounds, 1, nil, AScrollBar);
  Groups := AGroups;
  I := 0;
  while AGroups <> nil do
  begin
    AGroups := AGroups^.Next;
    Inc(I);
  end;
  SetRange(I);
  ComponentType:=ctElementGroupList;
end;

constructor TElementGroupList.Load(var S: TStream);

function ReadItems: PElementItem;
var
  Itms:  PElementItem;
  CurItm: ^PElementItem;
  Count, I: Integer;
begin
  S.Read(Count, SizeOf(Integer));
  Itms := nil;
  CurItm := @Itms;
  for I := 1 to Count do
  begin
    New(CurItm^);
    FillChar(CurItm^^,SizeOf(TElementItem),0);
    CurItm^^.Name := S.ReadStr;
    CurItm := @CurItm^^.Next;
  end;
  CurItm^ := nil;
  ReadItems := Itms;
end;

function ReadGroups: PElementGroup;
var
  Grps:  PElementGroup;
  CurGrp: ^PElementGroup;
  Count, I: Integer;
begin
  S.Read(Count, SizeOf(Integer));
  Grps := nil;
  CurGrp := @Grps;
  for I := 1 to Count do
  begin
    New(CurGrp^);
    FillChar(CurGrp^^, SizeOf(TElementGroup), 0);
    with CurGrp^^ do
    begin
      Name := S.ReadStr;
      Items := ReadItems;
    end;
    CurGrp := @CurGrp^^.Next;
  end;
  CurGrp^ := nil;
  ReadGroups := Grps;
end;

begin
  TListViewer.Load(S);
  Groups := ReadGroups;
end;

destructor TElementGroupList.Done;

procedure FreeItems(CurITem: PElementItem);
var
  P: PElementItem;
begin
  while CurItem <> nil do
  begin
    P := CurItem;
    DisposeStr(CurItem^.Name);
    CurItem := CurItem^.Next;
    Dispose(P);
  end;
end;

procedure FreeGroups(CurGroup: PElementGroup);
var
  P: PElementGroup;
begin
  while CurGroup <> nil do
  begin
    P := CurGroup;
    FreeItems(CurGroup^.Items);
    DisposeStr(CurGroup^.Name);
    CurGroup := CurGroup^.Next;
    Dispose(P);
  end
end;

begin
  TListViewer.Done;
  FreeGroups(Groups);
end;

procedure TElementGroupList.FocusItem(Item: Integer);
var
  CurGroup: PElementGroup;
begin
  TListViewer.FocusItem(Item);
  CurGroup := Groups;
  while Item > 0 do
  begin
    CurGroup := CurGroup^.Next;
    Dec(Item);
  end;
  Message(Owner, evBroadcast, cmNewElementItem, CurGroup);
end;

function TElementGroupList.GetText(Item: Integer; MaxLen: Integer): String;
var
  CurGroup: PElementGroup;
  I: Integer;
begin
  CurGroup := Groups;
  while Item > 0 do
  begin
    CurGroup := CurGroup^.Next;
    Dec(Item);
  end;
  GetText := CurGroup^.Name^;
end;

procedure TElementGroupList.Store(var S: TStream);

procedure WriteItems(Items: PElementItem);
var
  CurItm: PElementItem;
  Count: Integer;
begin
  Count := 0;
  CurItm := Items;
  while CurItm <> nil do
  begin
    CurItm := CurItm^.Next;
    Inc(Count);
  end;
  S.Write(Count, SizeOf(Integer));
  CurItm := Items;
  while CurItm <> nil do
  begin
    S.WriteStr(CurItm^.Name);
    CurItm := CurItm^.Next;
  end;
end;

procedure WriteGroups(Groups: PElementGroup);
var
  CurGrp: PElementGroup;
  Count: Integer;
begin
  Count := 0;
  CurGrp := Groups;
  while CurGrp <> nil do
  begin
    CurGrp := CurGrp^.Next;
    Inc(Count);
  end;
  S.Write(Count, SizeOf(Integer));
  CurGrp := Groups;
  while CurGrp <> nil do
  begin
    with CurGrp^ do
    begin
      S.WriteStr(Name);
      WriteItems(Items);
    end;
    CurGrp := CurGrp^.Next;
  end;
end;

begin
  TListViewer.Store(S);
  WriteGroups(Groups);
end;

procedure TElementGroupList.Awaken;
var
  CurGroup: PElementGroup;
  Item:integer;
begin
  Item:= Focused;
  CurGroup := Groups;
  while Item > 0 do
  begin
    CurGroup := CurGroup^.Next;
    Dec(Item);
  end;
  Message(Owner, evBroadcast, cmNewElementItem, CurGroup);
end;

procedure TElementGroupList.HandleEvent(var Event: TEvent);
begin
  TListViewer.HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmNewElementIndex) then
  begin
    SetGroupIndex(Focused, Event.InfoByte);
    ClearEvent(Event);
  end;
end;

procedure TElementGroupList.SetGroupIndex(GroupNum, ItemNum: Byte);
var
  Group: PElementGroup;
begin
  Group := GetGroup(GroupNum);
  if Group <> nil then
    Group^.Index := ItemNum;
end;

function TElementGroupList.GetGroupIndex(GroupNum: Byte): byte;
var
  Group: PElementGroup;
begin
  Group := GetGroup(GroupNum);
  if Group <> nil then
    GetGroupIndex := Group^.Index
  else
    GetGroupIndex := 0;
end;

function TElementGroupList.GetGroup(GroupNum: Byte): PElementGroup;
var
  Group: PElementGroup;
begin
  Group := Groups;
  while GroupNum > 0 do
  begin
    Group := Group^.Next;
    Dec(GroupNum);
  end;
  GetGroup := Group;
end;

function TElementGroupList.GetNumGroups: byte;
var
  Index: byte;
  Group: PElementGroup;
begin
  Index := 0;
  Group := Groups;
  while Group <> nil do
  begin
    Inc(Index);
    Group := Group^.Next;
  end;
  GetNumGroups := Index;
end;

{ TElementItemList }

constructor TElementItemList.Init(var Bounds: TRect; AScrollBar: PScrollBar;
  AItems: PElementItem);
var
  I: Integer;
begin
  TListViewer.Init(Bounds, 1, nil, AScrollBar);
  EventMask := EventMask or evBroadcast;
  Items := AItems;
  I := 0;
  while AItems <> nil do
  begin
    AItems := AItems^.Next;
    Inc(I);
  end;
  SetRange(I);
  ComponentType:=ctElementItemList;
end;

procedure TElementItemList.FocusItem(Item: Integer);
var
  CurItem: PElementItem;
begin
  TListViewer.FocusItem(Item);
  CurItem := Items;
  Message(Owner, evBroadcast, cmNewElementIndex, Pointer(Item));
  while Item > 0 do
  begin
    CurItem := CurItem^.Next;
    Dec(Item);
  end;
end;

function TElementItemList.GetText(Item: Integer; MaxLen: Integer): String;
var
  CurItem: PElementItem;
begin
  CurItem := Items;
  while Item > 0 do
  begin
    CurItem := CurItem^.Next;
    Dec(Item);
  end;
  GetText := CurItem^.Name^;
end;

procedure TElementItemList.HandleEvent(var Event: TEvent);
var
  CurItem: PElementItem;
  Group: PElementGroup;
  I: Integer;
begin
  if (Event.What = evMouseDown) and (Event.Double) then
  begin
    Event.What := evCommand;
    Event.Command := cmOK;
    PutEvent(Event);
    ClearEvent(Event);
  end;
  TListViewer.HandleEvent(Event);
  if Event.What = evBroadcast then
  case Event.Command of
    cmNewElementItem:
      begin
        Group := Event.InfoPtr;
        Items := Group^.Items;
        CurItem := Items;
        I := 0;
        while CurItem <> nil do
        begin
          CurItem := CurItem^.Next;
          Inc(I);
        end;
        SetRange(I);
        FocusItem(Group^.Index);
        DrawView;
      end;
  end;
end;

function TElementItemList.DataSize:Word;
begin
  DataSize:=SizeOf(String);
end;

procedure TElementItemList.GetData(var Rec);
begin
 String(Rec):=GetText(Focused, 255);
end;

procedure TElementItemList.SetData(var Rec);
begin
end;

{ -- Element list building routines -- }

function ElementItem(const Name: String; Next: PElementItem): PElementItem;
var
  Item: PElementItem;
begin
  New(Item);
  Item^.Name := NewStr(Name);
  Item^.Next := Next;
  ElementItem := Item;
end;

function ElementGroup(const Name: String; Items: PElementItem;
  Next: PElementGroup): PElementGroup;
var
  Group: PElementGroup;
begin
  New(Group);
  Group^.Name := NewStr(Name);
  Group^.Items := Items;
  Group^.Next := Next;
  Group^.Index:=0;
  ElementGroup := Group;
end;

procedure RegisterEl_Lists;
begin
  RegisterType(RElementGroupList);
  RegisterType(RElementItemList);
end;
end.
