{************************************************}
{                                                }
{   Turbo Pascal 7.0                             }
{   FreePacal                                    }
{   Delphi 4                                     }
{   Windows list dialog                          }
{   Copyright (c) 1995,96,97 by Alexs            }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}

{$I Define.inc}

unit ListWind;

interface

uses Objects, Drivers, Views, App, Dialogs, AddCntrl, Register;

type
  PWindowBox = ^TWindowBox;
  TWindowBox = object(TOkListBox)
    destructor Done; virtual;
    function GetText(Item: Integer; MaxLen: Integer): String; virtual;
  end;

type
  PListDialog = ^TListDialog;
  TListDialog = object(TDialog)
    WindowsList: PWindowBox;
    constructor Init;
    function Execute: Word; virtual;
    function GetSelectedWindow: PWindow;
    function GetWindowsList: PCollection;
    procedure HandleEvent(var Event: TEvent); virtual;
  end;

procedure WindowListDlg;
implementation

{ TWindowBox }

destructor TWindowBox.Done;
begin
  if List <> nil then
  begin
     List^.DeleteAll;
     Dispose(List, Done);
  end;
  TListBox.Done;
end;

function TWindowBox.GetText(Item: Integer; MaxLen: Integer): String;
var
  P: PWindow;
begin
  P := List^.At(Item);
  if P <> nil then  GetText := Copy(P^.GetTitle(255),1,MaxLen)
  else GetText := '';
end;

{ TWindowBoxDialog }

constructor TListDialog.Init;
var
  R, P: TRect;
  PS: PScrollBar;
begin
  R.Assign(0, 0, 51, 14);
  inherited Init(R, 'Windows List');
  Options := Options or ofCentered;
  GetExtent(R);
  R.Grow(-1,-1);
  inc(R.A.Y);
  R.B.X:=R.A.X+37;
  P.Assign(R.B.X, R.A.Y, R.B.X+1, R.B.Y);
   PS := New(PScrollBar, Init(P));
   Insert(PS);
  WindowsList := New(PWindowBox, Init(R, 1, PS));
    WindowsList^.NewList(GetWindowsList);
    if WindowsList^.List^.Count>1 then WindowsList^.FocusItem(1);
    Insert(WindowsList);
  R.Assign(1,1,10,2);
    Insert(New(PLabel, Init(R, '~W~indows', WindowsList)));
  R.Assign(39, 2, 49, 4);
    Insert(New(PButton, Init(R, '~O~k', cmOk, bfDefault)));
  R.Move(0, 3);
    Insert(New(PButton, Init(R, '~D~elete', cmDelWindow, bfNormal+bfBroadcast)));
  R.Move(0, 3);
    Insert(New(PButton, Init(R, 'Cancel', cmCancel, bfNormal)));
  R.Move(0, 3);
    Insert(New(PButton, Init(R, '~H~elp', cmHelp, bfNormal)));
  SelectNext(False);
end;

function TListDialog.Execute: Word;
var R:word;
    W:PView;
begin
  if (WindowsList = nil) or (WindowsList^.Range = 0) then
    DisableCommands([cmDelWindow]);
  R:=TDialog.Execute;
  Execute := R;
  if R=cmOk then
  begin
    W:=GetSelectedWindow;
    if W<>nil then W^.Select;
  end;
end;

function TListDialog.GetSelectedWindow: PWindow;
begin
  GetSelectedWindow := nil;
  if (WindowsList <> nil) and (WindowsList^.Range > 0) then
    GetSelectedWindow := WindowsList^.List^.At(WindowsList^.Focused);
end;

function TListDialog.GetWindowsList: PCollection;
var
  C: PCollection;
  Temp:PView;
procedure CollectWindows(P: PView);
begin
  if (P <> nil) and (P^.Options and (ofSelectable + ofTopSelect)<>0) and
     (P^.State and sfVisible<>0) then
      C^.Insert(P);
end;
begin
  C := New(PCollection, Init(10, 5));
  if (Desktop <> nil) and (C <> nil) and (Desktop^.Last<>nil) then
  with Desktop^ do
  begin
    Temp:=Last;
    repeat
      Temp:=Temp^.Next;
      if (Temp<>nil) then CollectWindows(Temp);
    until (Temp=Last) or (Temp=nil);
  end;
  GetWindowsList := C;
end;

procedure TListDialog.HandleEvent(var Event: TEvent);

procedure DeleteWindow;
var W:PWindow;
begin
  W := GetSelectedWindow;
  if W <> nil then
  begin
    WindowsList^.List^.Delete(W);
    WindowsList^.SetRange(WindowsList^.List^.Count);
    WindowsList^.DrawView;
    W^.Free;
    EnableCommands([cmClose]);
  end;
  if WindowsList^.Range = 0 then DisableCommands([cmDelWindow]);
end;

begin
  inherited HandleEvent(Event);
  case Event.What of
    evBroadcast: if Event.Command=cmDelWindow then DeleteWindow else exit;
    evKeyDown: if Event.KeyCode=kbDel then DeleteWindow else exit;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure WindowListDlg;
var D:PDialog;
begin
  D:=New(PListDialog, Init);
  Application^.ExecView(D);
  D^.Free;
end;

end.
