{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       Printer manager                                 }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}
unit Prn_Mng;
interface

uses Objects, Views, Dialogs, Drivers, Ext_Prn, Register;


type
   PPrinterListView = ^TPrinterListView;
   TPrinterListView = object(TListViewer)
     function GetText(Item: Integer; MaxLen: Integer): String; virtual;
   end;
type
   PPrnManager = ^TPrnManager;
   TPrnManager = object(TDialog)
     List:PPrinterListView;
     constructor Init;
     destructor Done;virtual;
     procedure Draw;virtual;
     procedure HandleEvent(var Event:TEvent);virtual;
     procedure Update;
   end;

const
   PrnManager:PPrnManager = nil;
implementation
uses Menus, App, Ext_Math, ResTools;

   { TPrinterListView }
function TPrinterListView.GetText(Item: Integer; MaxLen: Integer): String;
var Rec:PPrinterRec;
    S:string;
    F:real;
begin
  Rec:=PrinterManager^.At(Item);
  F:=Round((Rec^.Position/Rec^.Size) * 1000)/10;
  if PrinterManager^.Stop then S:='⠭' else S:='⠥';
  if Item<>0 then S:='';
  GetText:=Rec^.Name+' - '+S+' - '+FloatToStr(F)+'%';
end;

   { TPrnManager }
constructor TPrnManager.Init;
var
  R:TRect;
  ScrollBar:PScrollBar;
begin
  R.Assign(  7,  1, 58, 17);
  inherited Init(R,'Printer manager');
  PrnManager:=@Self;
  Palette:=dpBlueDialog;
  GrowMode := gfGrowAll + gfGrowRel;
  Flags := wfMove + wfGrow + wfClose + wfZoom;
  Options := Options or ofTileable or ofCentered;
  Number:=GetWindowNumber;
  GetExtent(R);
    R.Grow(0,-1);
    R.A.X:=R.B.X-1;
    R.A.Y:=R.A.Y+1;
    ScrollBar:=New(PScrollBar, Init( R));
    Insert(ScrollBar);
  GetExtent(R);
  R.Grow(-1,-1);
  Inc(R.A.Y);
    List:=New(PPrinterListView, Init(R, 1, nil, ScrollBar));
    List^.GrowMode := gfGrowHiX + gfGrowHiY;
    Insert(List);
  GetExtent(R);
    R.Grow(-1,-1);
    R.B.Y:=R.A.Y+1;
    Insert(New(PMenuBar, Init(R, NewMenu(
      NewItem('~~⠭', '', kbIns, cmStopPrint, hcNoContext,
      NewItem('~~த', '', kbNoKey, cmContinuePrint, hcNoContext,
      NewItem('~~', '', kbDel, cmDelItem, hcNoContext,
      NewItem('~~⢠', '', kbNoKey, cmError, hcNoContext,
    nil))))))));
  Update;
end;

destructor TPrnManager.Done;
begin
  inherited Done;
  PrnManager:=nil;
end;

procedure TPrnManager.HandleEvent(var Event:TEvent);

procedure DelItem;
begin
  if PrinterManager^.Count=0 then exit;
  PrinterManager^.AtFree(List^.Focused);
  Update;
end;

procedure ContinuePrint;
begin
  if PrinterManager^.Stop then PrinterManager^.Stop:=false;
end;

procedure StopPrint;
begin
  if not PrinterManager^.Stop then PrinterManager^.Stop:=true;
end;

begin
  inherited HandleEvent(Event);
  if PrinterManager=nil then exit;
  case Event.What of
    evBroadcast:
    begin
      case Event.Command of
        cmIdle:if (not PrinterManager^.Stop) and (PrinterManager^.Count<>0) then Update;
      end;
      exit;
    end;
    evCommand:
    case Event.Command of
      cmDelItem:DelItem;
      cmContinuePrint:ContinuePrint;
      cmStopPrint:StopPrint;
    else
      exit;
    end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TPrnManager.Update;
begin
   if PrinterManager<>nil then List^.SetRange(PrinterManager^.Count)
   else List^.SetRange(0);
   ReDraw;
end;

procedure TPrnManager.Draw;
begin
  if (PrinterManager<>nil) and
     (List^.Range<>PrinterManager^.Count) then
     List^.SetRange(PrinterManager^.Count);
  inherited Draw;
end;

end.
