uses Objects, DAE, Views, App, EditWind, Menus, Drivers, Register,
     MsgBox, StdDlg, Ext_Str, Ext_Dos, Ext_Crt, Dialogs, Ext_Sys,
     Validate, Ext_Math, Std_Str;
const
  CStatusText = #2;
type
  PStatusText = ^TStatusText;
  TStatusText = object(TStaticText)
    function GetPalette: PPalette;virtual;
  end;

function TStatusText.GetPalette: PPalette;
const
  P: String[Length(CStatusText)] = CStatusText;
begin
  GetPalette := @P;
end;

type
  TAliasEditor = object(TApplication)
    FileName:FNameStr;
    Status:PStatusText;
    constructor Init;
    procedure InitMenuBar; virtual;
    procedure InitStatusLine; virtual;
    function Valid(Command: Word): Boolean; virtual;
    function Save:boolean;
    function SaveAs:boolean;
    procedure Open;
    procedure HandleEvent(var Event:TEvent);virtual;
    procedure UpdateTitle;
  end;

constructor TAliasEditor.Init;
var R:TRect;
begin
  inherited Init;
  GetExtent(R);
  R.A.X:=R.B.X-40;
  R.B.Y:=R.A.Y+1;
  Status:=Insert(New(PStatusText, Init(R,'')));
  FileName:=DaeLibCfgFileName;
  DisableCommands([cmEditAlias,cmDeleteAlias]);
  UpdateTitle;
end;

procedure TAliasEditor.HandleEvent(var Event:TEvent);
begin
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
    case Event.Command of
      cmSave:Save;
      cmSaveAs:SaveAs;
      cmOpen:Open;
    else
      exit;
    end;
  else
    exit;
  end;
  ClearEvent(Event);
end;

procedure TAliasEditor.UpdateTitle;
begin
  AssignStr(Status^.Text,FileName);
  Status^.DrawView;
end;

procedure TAliasEditor.InitMenuBar;
var
  R: TRect;
begin
  GetExtent(R);
  R.B.Y := R.A.Y + 1;
  MenuBar := New(PMenuBar, Init(R, NewMenu(
    NewSubMenu('~F~ile', hcNoContext, NewMenu(
      StdFileMenuItems(
      nil)),
    NewSubMenu('~E~dit', hcNoContext, NewMenu(
      NewItem('~N~ew alias...', 'Ins', kbIns, cmNewAlias, hcOpen,
{      NewItem('~E~dit alias...', '', kbF3, cmEditAlias, hcOpen,}
      NewItem('~D~elete alias...', 'Del', kbDel, cmDeleteAlias, hcOpen,
      nil))),
    nil)))));
end;

procedure TAliasEditor.InitStatusLine;
var
  R: TRect;
begin
  GetExtent(R);
  R.A.Y := R.B.Y - 1;
  New(StatusLine, Init(R,
    NewStatusDef(0, $FFFF,
      NewStatusKey('~Alt+X~ E~x~it', kbAltX, cmQuit,
      NewStatusKey('~F2~ Save', kbF2, cmSave,
      NewStatusKey('~F3~ Open', kbF3, cmOpen,
      NewStatusKey('~F10~ Menu', kbF10, cmMenu,
      NewStatusKey('', kbCtrlF5, cmResize,
      nil))))),
    nil)));
end;

procedure TAliasEditor.Open;
begin
  ExecuteDialog(New(PFileDialog, Init('*.cfg',
         'Open file', '~N~ame', fdOpenButton, 101)), nil);
end;

function TAliasEditor.Save:boolean;
begin
  if FileName<>'' then
  begin
     Session^.StoreCfgFile(FileName);
     Save:=true;
  end
  else Save:=SaveAs;
end;

function TAliasEditor.SaveAs:boolean;
var TempName:FNameStr;
    Re:word;
begin
  TempName:=FileName;
  Re:=ExecuteDialog(New(PFileDialog, Init('*.cfg',
         'Save file as', '~N~ame', fdOkButton, 101)), @TempName);
  if (Re<>cmCancel) and (stUpCase(TempName)<>stUpCase(FileName)) then
  begin
    if FileExists(TempName) then
      if MessageBox('File %s already exists. Overwrite?',@TempName,
         mfInsertInApp+mfConfirmation+mfYesButton+mfNoButton)<>cmYes then
                Re:=cmCancel;
  end;
  if Re<>cmCancel then
  begin
    FileName:=TempName;
    Session^.StoreCfgFile(FileName)
  end;
  SaveAs:=Re=cmOk;
  UpdateTitle;
end;

function TAliasEditor.Valid(Command: Word): Boolean;
var B:boolean;
begin
  B:=inherited Valid(Command);
  if (Command = cmQuit) and Modified then
  case MessageBox('Config has been modified. Save?', nil,
                  mfConfirmation+mfYesNoCancel) of
    cmYes:B:=Save;
    cmCancel:B:=false;
  end;
  Valid:=B;
end;

var
  AllEdit:TAliasEditor;
  W:PWindow;
begin
  ClrScr;
  InitDataSets;
  AllEdit.Init;
  W:=New(PAliasEditWindow, Init);
  W^.Flags:=0;
  AllEdit.InsertWindow(W);
  AllEdit.Run;
  AllEdit.Done;
  DoneDataSets;
end.
